/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mips.h"
#include "pic32.h"

#define HFREQ_LIMIT	627	/* 20MHz cycles */
#define HSYNC_LIMIT	16	/* bytes or pixels */
#define LINE_LIMIT	160	/* bytes or pixels */

#define VSYNC_START	1	/* horizontal lines, front porch end */
#define VISIBLE_START	4	/* horizontal lines */
#define VBP_START	516	/* horizontal lines, back porch start */
#define VBP_END		531	/* horizontal lines, back porch end */

/* Disable JTAG functionality on pins. */

.section .devcfg0, "a"
.word 0xfffffffb	/* DEVCFG0<2> = JTAGEN = 0 */

/*
Set the oscillator to be the FRC oscillator with PLL, with peripheral clock
divided by 1, HS oscillator mode selected (for PLL), and FRCDIV+PLL selected.
With a system clock of 20MHz, the peripheral clock is therefore 20MHz.

The watchdog timer (FWDTEN) is also disabled.
*/

.section .devcfg1, "a"
.word 0xff7fcff9	/* DEVCFG1<23> = FWDTEN = 0; DEVCFG1<13:12> = FPBDIV<2:0> = 0; DEVCFG1<2:0> = FNOSC<2:0> = 001 */

/*
Set the FRC oscillator PLL function with an input division of 4, an output
division of 2, a multiplication of 20, yielding a multiplication of 2.5.
This should take the FRC at 8MHz and produce a system clock of 20MHz.
*/

.section .devcfg2, "a"
.word 0xfff9ffdb	/* DEVCFG2<18:16> = FPLLODIV<2:0> = 1; DEVCFG2<6:4> = FPLLMUL<2:0> = 101; DEVCFG2<2:0> = FPLLIDIV<2:0> = 011 */

.text
.globl _start

_start:
	/*
	Configure RAM.
	See: http://microchipdeveloper.com/32bit:mx-arch-exceptions-processor-initialization
	*/

	la $v0, BMXCON
	li $v1, (1 << 6)	/* BMXCON<6> = BMXWSDRM = 0 */
	sw $v1, CLR($v0)

	/* Enable caching. */

	li $v0, CONFIG_CM_CACHABLE_NONCOHERENT
	mtc0 $v0, CP0_CONFIG
	nop

	/* Get the RAM size. */

	la $v0, BMXDRMSZ
	lw $v0, 0($v0)

	/* Initialise the stack pointer. */

	li $v1, KSEG0_BASE
	addu $sp, $v0, $v1	/* sp = KSEG0_BASE + RAM size */

	/* Initialise the globals pointer. */

	lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
	ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)

	/* Set pins for output. */

	jal init_pins
	nop

	jal init_oc_pins
	nop

	/* Initialise the status register. */

	jal init_interrupts
	nop

	/* Initialise timer. */

	jal init_timer2
	nop

	/* Initialise PMP. */

	jal init_pmp
	nop

	/* Initialise DMA. */

	jal init_dma
	nop

	/* Initialise OC2. */

	jal init_oc2
	nop

	/* Initialise framebuffer. */

	jal init_framebuffer
	nop

	/* Enable interrupts and loop. */

	jal enable_interrupts
	nop

	jal handle_error_level
	nop

	/* Main program. */

	li $a1, 20000000	/* counter = 20000000 */

	/* Initialise the display state. */

	li $s0, 0		/* line counter */
	la $s1, vfp_active	/* current event */
	move $s2, $zero		/* line address */

	/* Monitoring loop. */
loop:
	addiu $a1, $a1, -1	/* counter -= 1 */
	bnez $a1, loop		/* until counter == 0 */
	nop

	li $a1, 20000000	/* counter = 20000000 */

	la $t0, PORTB
	li $t1, (1 << 10)	/* PORTB<10> = RB10 */
	sw $t1, INV($t0)

_next:
	j loop
	nop



init_pins:
	/* DEVCFG0<2> needs setting to 0 before the program is run. */

	la $v0, CFGCON
	li $v1, (1 << 3)	/* CFGCON<3> = JTAGEN = 0 */
	sw $v1, CLR($v0)

init_outputs:
	/* Remove analogue features from pins. */

	la $v0, ANSELA
	sw $zero, 0($v0)	/* ANSELA = 0 */
	la $v0, ANSELB
	sw $zero, 0($v0)	/* ANSELB = 0 */

	la $v0, TRISA
	sw $zero, 0($v0)
	la $v0, TRISB
	sw $zero, 0($v0)

	jr $ra
	nop



/* Interrupt servicing. */

.org 0x200

interrupt_handler:

	/* Check for a timer interrupt condition. */

	la $v0, IFS0
	lw $v1, 0($v0)
	andi $v1, $v1, (1 << 9)	/* T2IF */
	beqz $v1, irq_pmp
	nop

	/* Increment the line counter. */

	addiu $s0, $s0, 1

	/* Jump to the event handler. */

	jalr $s1
	nop

irq_clear_timer:

	/* Clear the timer interrupt condition. */

	la $v0, IFS0
	li $v1, (1 << 9)	/* IFS0<9> = T2IF = 0 */
	sw $v1, CLR($v0)

irq_pmp:

	/* Check for a PMP interrupt condition. */

	la $v0, IFS1
	lw $v1, 0($v0)
	li $t8, (1 << 16)	/* PMPIF */
	and $v1, $v1, $t8
	beqz $v1, irq_exit
	nop

irq_clear_pmp:

	/* Clear the PMP interrupt condition. */

	la $v0, IFS1
	li $v1, (1 << 16)	/* IFS1<16> = PMPIF = 0 */
	sw $v1, CLR($v0)

irq_exit:
	eret
	nop



/* Utilities. */

handle_error_level:
	mfc0 $t3, CP0_STATUS
	li $t4, ~(STATUS_ERL | STATUS_EXL)
	and $t3, $t3, $t4
	mtc0 $t3, CP0_STATUS
	jr $ra
	nop

enable_interrupts:
	mfc0 $t3, CP0_STATUS
	nop
	ori $t3, $t3, STATUS_IRQ | STATUS_IE
	mtc0 $t3, CP0_STATUS
	jr $ra
	nop

init_interrupts:
	mfc0 $t3, CP0_STATUS
	li $t4, ~STATUS_BEV
	and $t3, $t3, $t4
	mtc0 $t3, CP0_STATUS	/* CP0_STATUS &= ~STATUS_BEV (use non-bootloader vectors) */

	la $t3, _start
	mtc0 $t3, CP0_EBASE	/* EBASE = _start */

	li $t3, CAUSE_IV	/* IV = 1 (use EBASE+0x200 for interrupts) */
	mtc0 $t3, CP0_CAUSE

	jr $ra
	nop



/* Event routines. */

/* Start of, and within, the vertical front porch. */

vfp_start:
	/* Clear vsync. */

	la $v0, PORTB
	la $v1, (1 << 11)	/* PORTB<11> = RB11 */
	sw $v1, CLR($v0)

	/* Enter the active region. */

	la $s1, vfp_active

vfp_active:
	/* Test for vsync. */

	subu $v0, $s0, VSYNC_START
	bnez $v0, _vfp_active_ret
	nop

	/* Start the vsync region. */

	la $s1, vsync_start

_vfp_active_ret:
	jr $ra
	nop



/* Start of, and within, vsync. */

vsync_start:
	/* Set vsync. */

	la $v0, PORTB
	la $v1, (1 << 11)	/* PORTB<11> = RB11 */
	sw $v1, SET($v0)

	/* Enter the active region. */

	la $s1, vsync_active

vsync_active:
	/* Test for visible region. */

	subu $v0, $s0, VISIBLE_START
	bnez $v0, _vsync_active_ret
	nop

	/* Start the visible region. */

	la $s1, visible_start

	/* Enable OC2. */

	la $v0, OC2CON
	li $v1, (1 << 15)
	sw $v1, SET($v0)

_vsync_active_ret:
	jr $ra
	nop



/* Start of, and within, the visible period. */

visible_start:
	/* Clear vsync. */

	la $v0, PORTB
	la $v1, (1 << 11)	/* PORTB<11> = RB11 */
	sw $v1, CLR($v0)

	/* Enter the active region. */

	la $s1, visible_active

visible_active:
	/* Test for back porch. */

	subu $v0, $s0, VBP_START
	bnez $v0, _visible_active_ret
	nop

	/* Start the back porch. */

	la $s1, vbp_active

	/* Disable OC2. */

	la $v0, OC2CON
	li $v1, (1 << 15)
	sw $v1, CLR($v0)

_visible_active_ret:
	jr $ra
	nop



/* Within the vertical back porch. */

vbp_active:
	/* Test for front porch. */

	subu $v0, $s0, VBP_END
	bnez $v0, _vbp_active_ret
	nop

	/* Start the front porch. */

	li $s0, 0
	la $s1, vfp_start

_vbp_active_ret:
	jr $ra
	nop



/* Initialisation routines. */

init_timer2:

	/* Initialise Timer2 interrupt. */

	la $v0, T2CON
	sw $zero, 0($v0)	/* T2CON = 0 */
	nop

	la $v0, TMR2
	sw $zero, 0($v0)	/* TMR2 = 0 */

	la $v0, PR2
	li $v1, HFREQ_LIMIT
	sw $v1, 0($v0)		/* PR2 = HFREQ_LIMIT */

	/* Initialise Timer2 interrupt. */

	la $v0, IFS0
	li $v1, (1 << 9)
	sw $v1, CLR($v0)	/* T2IF = 0 */

	la $v0, IPC2
	li $v1, 0b11111
	sw $v1, CLR($v0)	/* T2IP, T2IS = 0 */

	la $v0, IPC2
	li $v1, 0b11111
	sw $v1, SET($v0)	/* T2IP = 7; T2IS = 3 */

	la $v0, IEC0
	li $v1, (1 << 9)
	sw $v1, SET($v0)	/* T2IE = 1 */

	/* Start timer. */

	la $v0, T2CON
	li $v1, (1 << 15)
	sw $v1, SET($v0)	/* ON = 1 */

	jr $ra
	nop



/*
Output compare initialisation.

Timer2 will be used to trigger two events: one initiating the hsync pulse, and
one terminating the pulse. Upon the termination of the pulse, an interrupt
condition will cause the line data to be transferred using DMA.

*/

init_oc2:
	/* Disable OC interrupts. */

	la $v0, IEC0
	li $v1, (1 << 12)	/* IEC0<12> = OC2IE = 0 */
	sw $v1, CLR($v0)

	la $v0, IFS0
	li $v1, (1 << 12)	/* IFS0<12> = OC2IF = 0 */
	sw $v1, CLR($v0)

	/* Initialise OC2. */

	la $v0, OC2CON
	li $v1, 0b101		/* OC2CON<2:0> = OCM<2:0> = 101 (dual compare, continuous pulse) */
	sw $v1, 0($v0)

	/* Set hsync start and end. */

	la $v0, OC2R
	li $v1, 16
	sw $v1, 0($v0)

	la $v0, OC2RS
	li $v1, 32
	sw $v1, 0($v0)

	/* OC2 is enabled elsewhere when it needs to operate. */

	jr $ra
	nop



init_oc_pins:
	/* Unlock the configuration register bits. */

	la $v0, SYSKEY
	sw $zero, 0($v0)
	li $v1, 0xAA996655
	sw $v1, 0($v0)
	li $v1, 0x556699AA
	sw $v1, 0($v0)

	la $v0, CFGCON
	lw $t8, 0($v0)
	li $v1, (1 << 13)	/* IOLOCK = 0 */
	sw $v1, CLR($v0)

	/* Map OC2 to RPB5. */

	la $v0, RPB5R
	li $v1, 0b0101		/* RPB5R<3:0> = 0101 (OC2) */
	sw $v1, 0($v0)

	la $v0, CFGCON
	sw $t8, 0($v0)

	/* Lock the oscillator control register again. */

	la $v0, SYSKEY
	li $v1, 0x33333333
	sw $v1, 0($v0)

	jr $ra
	nop



/* Parallel Master Port initialisation. */

init_pmp:
	/* Disable PMP interrupts. */

	la $v0, IEC1
	li $v1, (1 << 16)	/* IEC1<16> = PMPIE = 0 */
	sw $v1, CLR($v0)

	/* Initialise PMP.

	PMCON<12:11> = ADDRMUX<1:0> = 0; demultiplexed address and data
	PMCON<9> = PTWREN<0> = 0; no write pin
	PMCON<8> = PTRDEN<0> = 0; no read pin
	PMCON<7:6> = CSF<1:0> = 0; no chip select pins
	*/

	la $v0, PMCON
	sw $zero, 0($v0)

	/*
	PMMODE<14:13> = IRQM<1:0> = 1; interrupt after every read/write
	PMMODE<12:11> = INCM<1:0> = 0; no increment on every read/write
	PMMODE<10> = MODE16<0> = 0; 8-bit transfers
	PMMODE<9:8> = MODE<1:0> = 10; master mode 2
	PMMODE<5:2> = WAITM<3:0> = 00; single cycle read/write, no chip select wait cycles
	*/

	la $v0, PMMODE
	li $v1, 0x2200
	sw $v1, 0($v0)

	/* Free non-essential pins for general I/O. */

	la $v0, PMAEN
	sw $zero, 0($v0)

	la $v0, PMADDR
	sw $zero, 0($v0)

	la $v0, IFS1
	li $v1, (3 << 16)	/* IFS1<17:16> = PMPEIF, PMPIF = 0 */
	sw $v1, CLR($v0)

	/* Start PMP mode. */

	la $v0, PMCON
	li $v1, (1 << 15)	/* PMCON<15> = ON = 1 */
	sw $v1, SET($v0)

	jr $ra
	nop



/*
Direct Memory Access initialisation.

Write 160 pixels to the PMP for the line data. This is initiated by an output
compare interrupt.
*/

init_dma:
	/* Disable DMA interrupts. */

	la $v0, IEC1
	li $v1, (7 << 28)	/* IEC1<30:28> = DMA2IE, DMA1IE, DMA0IE = 0 */
	sw $v1, CLR($v0)

	/* Clear DMA interrupt flags. */

	la $v0, IFS1
	li $v1, (7 << 28)	/* IFS1<30:28> = DMA2IF, DMA1IF, DMA0IF = 0 */
	sw $v1, CLR($v0)

	/* Enable DMA. */

	la $v0, DMACON
	li $v1, (1 << 15)
	sw $v1, SET($v0)

	/*
	Initialise a line channel.
	The line channel will be channel 0 (x = 0).

	Once the hsync channel has completed a transfer, the line channel
	transfer is initiated.

	Specify a priority of 3:
	DCHxCON<1:0> = CHPRI<1:0> = 3

	Auto-enable the channels:
	DCHxCON<4> = CHAEN = 1
	*/

	la $v0, DCH0CON
	li $v1, 0b10011
	sw $v1, 0($v0)

	/*
	Initiate channel transfers when the initiating interrupt condition
	occurs:
	DCHxECON<15:8> = CHSIRQ<7:0> = output compare 2 interrupt
	*/

	la $v0, DCH0ECON
	li $v1, (12 << 8) | (1 << 4)
	sw $v1, 0($v0)

	/*
	The line channel has a cell size of 160 bytes:
	DCHxCSIZ<15:0> = CHCSIZ<15:0> = LINE_LIMIT
	*/

	la $v0, DCH0CSIZ
	li $v1, LINE_LIMIT
	sw $v1, 0($v0)

	/*
	Each source has a size identical to the cell size:
	DCHxSSIZ<15:0> = CHSSIZ<15:0> = n
	*/

	la $v0, DCH0SSIZ
	li $v1, LINE_LIMIT
	sw $v1, 0($v0)

	/*
	The source address is the physical address of the line data:
	DCHxSSA = line data physical address
	*/

	la $v0, DCH0SSA
	sw $zero, 0($v0)

	/*
	Each destination has a size of 1 byte:
	DCHxDSIZ<15:0> = CHDSIZ<15:0> = 1
	*/

	la $v0, DCH0DSIZ
	li $v1, 1
	sw $v1, 0($v0)

	/*
	For the line channel, the destination address is the physical address of
	PMDIN: DCHxDSA = physical(PMDIN)
	*/

	la $v0, DCH0DSA
	li $v1, PMDIN
	li $t8, KSEG1_BASE
	subu $v1, $v1, $t8
	sw $v1, 0($v0)

	la $v0, DCH0INT
	sw $zero, 0($v0)

	/* Enable channels. */

	la $v0, DCH0CON
	li $v1, 0b10000000
	sw $v1, SET($v0)

	jr $ra
	nop



/* Framebuffer initialisation. */

init_framebuffer:
	li $v0, KSEG0_BASE
	li $t8, 40 * 1024
	li $v1, 0xff

_init_fb_loop:
	sw $v1, 0($v0)
	addiu $v0, $v0, 4
	addiu $t8, $t8, -4
	bnez $t8, _init_fb_loop
	nop

	jr $ra
	nop
